;; ============================================================================
;; file:	msx_midi.asm              (Philips Music Module MIDI Framework)
;; version:     0.7a (March 1, 2008)
;;
;; description: This is a framework to use the MIDI functionality of the
;;              Philips Music Module (NMS1205). Only MIDI-IN is supported
;;              in this version. It should be easy to port to other MIDI
;;              interfaces.
;;
;;  This program is free software; you can redistribute it and/or modify
;;  it under the terms of the Lesser GNU General Public License as published by
;;  the Free Software Foundation; either version 2, or (at your option)
;;  any later version. See COPYING for more details.
;;
;; (C)2007,2008 by Vincent van Dam (vincentd@erg.verweg.com)
;; ============================================================================

	;; constants

msx_midi_command:	equ 0
msx_midi_dataout:	equ 1
msx_midi_status:	equ 4
msx_midi_datain:	equ 5

	;; -------------------------------------------------------------------
	;; buffers
	;; -------------------------------------------------------------------

_msx_midi_buffers_start:
msx_midi_in_buffer: ds  256         ; temporary midi in buffer
msx_midi_in_buffer_wr_pointer: dw 0 ; write pointer to data in midi in buffer
msx_midi_in_buffer_rd_pointer: dw 0 ; read pointer to data in midi in buffer
msx_midi_in_buffer_count: db 0      ; number of bytes ready in midi in buffer
msx_midi_out_buffer: ds 256         ; temporary midi out buffer
msx_midi_out_buffer_wr_pointer: dw 0; write pointer to data in midi out buffer
msx_midi_out_buffer_rd_pointer: dw 0; read pointer to data in midi out buffer
msx_midi_out_buffer_count: db  0    ; number of bytes ready in midi in buffer
msx_midi_out_buffer_rt: ds 256      ; temporary realtime midi out buffer
msx_midi_out_buffer_rt_wr_pointer: dw 0; write pointer to realtime midi out bfr
msx_midi_out_buffer_rt_rd_pointer: dw 0; read pointer to realtime midi out bffr
msx_midi_out_buffer_rt_count: db  0 ; number of bytes ready in midi in buffer
msx_midi_old_interrupt: ds  5       ; store for old interrupt handler
msx_midi_message_buffer: ds  3      ; buffer for non-sysex midi messages
msx_midi_message_count: db  0       ; counter to keeps expected no of bytes
msx_midi_message_pointer: dw  0     ; pointer to buffer
msx_midi_interrupt_flag: db  0	    ; set when interrupt should occur
msx_midi_running_status: db  0      ; last command triggered (running status)
msx_midi_handlers_noteon: ds  128   ; list of note-on handlers (max 31)
msx_midi_handlers_noteoff: ds  128  ; list of note-off handlers (max 31)
msx_midi_handlers_cc: ds  128       ; list of cc handlers (max 31)
msx_midi_handlers_aftertouch: ds 128; list of aftertouch handlers (max 31)
msx_midi_handlers_programm_change: ds 128; list programm change hndlrs (max 31)
msx_midi_handlers_channel_pressure: ds 128; list ch. pressure handlers (max 31)
msx_midi_handlers_pitchwheel: ds 128; list of pitchwheel handlers (max 31)
msx_midi_handlers_start: ds 32      ; list of start handlers (max 7)
msx_midi_handlers_stop:  ds 32      ; list of stop handlers (max 7)
msx_midi_handlers_clock: ds 32      ; list of cc handlers (max 7)
msx_midi_handler_midi_in: dw 0      ; handler when midi in data arrives
msx_midi_clock_type: db 0           ; 0 - internal clock, 1 - external clock 
_msx_midi_buffers_end:	
	
	;; --------------------------------------------------------------------
	;; msx_midi_init - initialize framework
	;; --------------------------------------------------------------------
	;; in:	[a ] = midi clock type (0 for internal, 1 for external clock)
	;; out:	-
	;; --------------------------------------------------------------------
	
msx_midi_init:
	di
	push  af
	call  msx_midi_midi_init
	;; clear buffers
	ld    hl,_msx_midi_buffers_start
	ld    de,_msx_midi_buffers_start+1
	ld    bc,_msx_midi_buffers_end-_msx_midi_buffers_start-1
	ld    (hl),0
	ldir
	;; midi in buffer
	ld    hl,msx_midi_in_buffer
	ld    (msx_midi_in_buffer_rd_pointer),hl
	ld    (msx_midi_in_buffer_wr_pointer),hl
	;; midi out buffer
	ld    hl,msx_midi_out_buffer
	ld    (msx_midi_out_buffer_wr_pointer),hl
	ld    (msx_midi_out_buffer_rd_pointer),hl
	;; midi out realtime message buffer
	ld    hl,msx_midi_out_buffer_rt
	ld    (msx_midi_out_buffer_rt_wr_pointer),hl
	ld    (msx_midi_out_buffer_rt_rd_pointer),hl
	;; midi message buffer
	ld    hl,msx_midi_message_buffer
	ld    (msx_midi_message_pointer),hl
	;; set midi clock type
	pop   af
	ld    (msx_midi_clock_type),a	
	;; init interrupt
	ld    hl,$fd9a
	ld    bc,5
	ld    de,msx_midi_old_interrupt
	ldir
	ld    a,$c3
	ld    ($fd9a),a
	ld    hl,msx_midi_interrupt
	ld    ($fd9b),hl
	ei
	ret

	;; --------------------------------------------------------------------
	;; msx_midi_exit - exit framework
	;; --------------------------------------------------------------------
	;; in:	-
	;; out:	-
	;; --------------------------------------------------------------------
	
msx_midi_exit:
	di
	;; restore interrupt handlers
	call  msx_midi_midi_reset
	ld    hl,msx_midi_old_interrupt
	ld    de,$fd9a
	ld    bc,5
	ldir
	;; clear buffers
	ld    hl,_msx_midi_buffers_start
	ld    de,_msx_midi_buffers_start+1
	ld    bc,_msx_midi_buffers_end-_msx_midi_buffers_start-1
	ld    (hl),0
	ldir
	ei
	ret

	;; --------------------------------------------------------------------
	;; msx_midi_add_clock_handler - add custom handler for midi clock
	;; --------------------------------------------------------------------
	;; in:	[hl] = address of clock handler to add
	;; out:	-
	;; --------------------------------------------------------------------
	
msx_midi_add_clock_handler:
	push  hl
	ld    de,0
	ld    hl,msx_midi_handlers_clock
	pop   bc
	call  msx_midi_add_handler
	ret

	;; --------------------------------------------------------------------
	;; msx_midi_del_clock_handler - remove custom handler for midi clock
	;; --------------------------------------------------------------------
	;; in:	[hl] = address of clock handler to remove
	;; out:	-
	;; --------------------------------------------------------------------
	
msx_midi_del_clock_handler:
	push  hl
	ld    hl,msx_midi_handlers_clock
	pop   bc
	call  msx_midi_del_handler
	ret	

	;; --------------------------------------------------------------------
	;; msx_midi_add_noteon_handler - add custom handler for note on
	;; --------------------------------------------------------------------
	;; in:	[hl] = address of note on handler to add
	;;      [ e] = midi channel (0 for all or 1-16)
	;; out:	-
	;; --------------------------------------------------------------------
	;; Handler will receive [a] = midi channel, [b] = velocity, [c] = note
	;; --------------------------------------------------------------------
	
msx_midi_add_noteon_handler:
	push  hl
	ld    d,0
	ld    hl,msx_midi_handlers_noteon
	pop   bc
	call  msx_midi_add_handler
	ret

	;; --------------------------------------------------------------------
	;; msx_midi_del_noteon_handler - remove custom handler for note on
	;; --------------------------------------------------------------------
	;; in:	[hl] = address of note on handler to remove
	;; out:	-
	;; --------------------------------------------------------------------
	
msx_midi_del_noteon_handler:
	push  hl
	ld    hl,msx_midi_handlers_noteon
	pop   bc
	call  msx_midi_del_handler
	ret	

	;; --------------------------------------------------------------------
	;; msx_midi_add_noteoff_handler - add custom handler for note off
	;; --------------------------------------------------------------------
	;; in:	[hl] = address of note off handler to add
	;;      [ e] = midi channel (0 for all or 1-16)
	;; out:	-
	;; --------------------------------------------------------------------
	;; Handler will receive [a] = midi channel, [b] = velocity, [c] = note
	;; --------------------------------------------------------------------
	
msx_midi_add_noteoff_handler:
	push  hl
	ld    d,0
	ld    hl,msx_midi_handlers_noteoff
	pop   bc
	call  msx_midi_add_handler
	ret

	;; --------------------------------------------------------------------
	;; msx_midi_del_noteoff_handler - remove custom handler for note off
	;; --------------------------------------------------------------------
	;; in:	[hl] = address of note off handler to remove
	;; out:	-
	;; --------------------------------------------------------------------
	
msx_midi_del_noteoff_handler:
	push  hl
	ld    hl,msx_midi_handlers_noteoff
	pop   bc
	call  msx_midi_del_handler
	ret	

	;; --------------------------------------------------------------------
	;; msx_midi_add_clock_handler - add custom handler for midi cc messages
	;; --------------------------------------------------------------------
	;; in:	[hl] = address of cc message handler to add
	;;      [d ] = cc id
	;;      [ e] = midi channel (0 for all or 1-16)
	;; out:	-
	;; --------------------------------------------------------------------
	;; Handler will receive [a] = midi channel, [b] = value, [c] = cc id
	;; --------------------------------------------------------------------
	
msx_midi_add_cc_handler:
	push  hl
	ld    hl,msx_midi_handlers_cc
	pop   bc
	call  msx_midi_add_handler
	ret

	;; --------------------------------------------------------------------
	;; msx_midi_del_cc_handler - remove custom handler for midi cc messages
	;; --------------------------------------------------------------------
	;; in:	[hl] = address of cc message handler to remove
	;; out:	-
	;; --------------------------------------------------------------------
	
msx_midi_del_cc_handler:
	push  hl
	ld    hl,msx_midi_handlers_cc
	pop   bc
	call  msx_midi_del_handler
	ret

	;; --------------------------------------------------------------------
	;; msx_midi_add_aftertouch_handler - add custom handler for aftertouch
	;; --------------------------------------------------------------------
	;; in:	[hl] = address of aftertouch handler to add
	;;      [ e] = midi channel (0 for all or 1-16)
	;; out:	-
	;; --------------------------------------------------------------------
	;; Handler will receive [a] = midi channel, [b] = pressure, [c] = note
	;; --------------------------------------------------------------------
	
msx_midi_add_aftertouch_handler:
	push  hl
	ld    d,0
	ld    hl,msx_midi_handlers_aftertouch
	pop   bc
	call  msx_midi_add_handler
	ret

	;; --------------------------------------------------------------------
	;; msx_midi_del_aftertouch_handler - remove custom handler aftertouch
	;; --------------------------------------------------------------------
	;; in:	[hl] = address of aftertouch handler to remove
	;; out:	-
	;; --------------------------------------------------------------------
	
msx_midi_del_aftertouch_handler:
	push  hl
	ld    hl,msx_midi_handlers_aftertouch
	pop   bc
	call  msx_midi_del_handler
	ret

	;; --------------------------------------------------------------------
	;; msx_midi_add_programm_change_handler - add custom handler prg change
	;; --------------------------------------------------------------------
	;; in:	[hl] = address of programm change handler to add
	;;      [ e] = midi channel (0 for all or 1-16)
	;; out:	-
	;; --------------------------------------------------------------------
	;; Handler will receive [a] = midi channel, [c] = programm
	;; --------------------------------------------------------------------
	
msx_midi_add_programm_change_handler:
	push  hl
	ld    d,0
	ld    hl,msx_midi_handlers_programm_change
	pop   bc
	call  msx_midi_add_handler
	ret

	;; --------------------------------------------------------------------
	;; msx_midi_del_programm_change_handler - remove custom handler prg chg
	;; --------------------------------------------------------------------
	;; in:	[hl] = address of programm change handler to remove
	;; out:	-
	;; --------------------------------------------------------------------
	
msx_midi_del_programm_change_handler:
	push  hl
	ld    hl,msx_midi_handlers_programm_change
	pop   bc
	call  msx_midi_del_handler
	ret	

	;; --------------------------------------------------------------------
	;; msx_midi_add_channel_pressure_handler - add custom handler ch press
	;; --------------------------------------------------------------------
	;; in:	[hl] = address of channel pressure handler to add
	;;      [ e] = midi channel (0 for all or 1-16)
	;; out:	-
	;; --------------------------------------------------------------------
	;; Handler will receive [a] = midi channel, [c] = pressure
	;; --------------------------------------------------------------------
	
msx_midi_add_channel_pressure_handler:
	push  hl
	ld    d,0
	ld    hl,msx_midi_handlers_channel_pressure
	pop   bc
	call  msx_midi_add_handler
	ret

	;; --------------------------------------------------------------------
	;; msx_midi_del_channel_pressure_handler - remove custom handler ch pr.
	;; --------------------------------------------------------------------
	;; in:	[hl] = address of channel pressure handler to remove
	;; out:	-
	;; --------------------------------------------------------------------
	
msx_midi_del_channel_pressure_handler:
	push  hl
	ld    hl,msx_midi_handlers_channel_pressure
	pop   bc
	call  msx_midi_del_handler
	ret	

	;; --------------------------------------------------------------------
	;; msx_midi_add_pitchweel_handler - add custom handler for pitchweel
	;; --------------------------------------------------------------------
	;; in:	[hl] = address of pitchwheel handler to add
	;;      [ e] = midi channel (0 for all or 1-16)
	;; out:	-
	;; --------------------------------------------------------------------
	;; Handler will receive [a] = midi channel, [bc] = pitchwheel value
	;; --------------------------------------------------------------------
	
msx_midi_add_pitchwheel_handler:
	push  hl
	ld    d,0
	ld    hl,msx_midi_handlers_pitchwheel
	pop   bc
	call  msx_midi_add_handler
	ret

	;; --------------------------------------------------------------------
	;; msx_midi_del_pitchwheel_handler - remove custom handler for pitchwhl
	;; --------------------------------------------------------------------
	;; in:	[hl] = address of pitchwheel handler to remove
	;; out:	-
	;; --------------------------------------------------------------------
	
msx_midi_del_pitchwheel_handler:
	push  hl
	ld    hl,msx_midi_handlers_pitchwheel
	pop   bc
	call  msx_midi_del_handler
	ret

	;; --------------------------------------------------------------------
	;; msx_midi_add_start_handler - add custom handler for midi start
	;; --------------------------------------------------------------------
	;; in:	[hl] = address of midi start handler to add
	;; out:	-
	;; --------------------------------------------------------------------
	
msx_midi_add_start_handler:
	push  hl
	ld    de,0
	ld    hl,msx_midi_handlers_start
	pop   bc
	call  msx_midi_add_handler
	ret

	;; --------------------------------------------------------------------
	;; msx_midi_del_start_handler - remove custom handler for midi start
	;; --------------------------------------------------------------------
	;; in:	[hl] = address of midi start handler to remove
	;; out:	-
	;; --------------------------------------------------------------------
	
msx_midi_del_start_handler:
	push  hl
	ld    hl,msx_midi_handlers_start
	pop   bc
	call  msx_midi_del_handler
	ret

	;; --------------------------------------------------------------------
	;; msx_midi_add_stop_handler - add custom handler for midi stop
	;; --------------------------------------------------------------------
	;; in:	[hl] = address of midi start handler to add
	;; out:	-
	;; --------------------------------------------------------------------
	
msx_midi_add_stop_handler:
	push  hl
	ld    de,0
	ld    hl,msx_midi_handlers_stop
	pop   bc
	call  msx_midi_add_handler
	ret

	;; --------------------------------------------------------------------
	;; msx_midi_del_stop_handler - remove custom handler for midi stop
	;; --------------------------------------------------------------------
	;; in:	[hl] = address of midi start handler to remove
	;; out:	-
	;; --------------------------------------------------------------------
	
msx_midi_del_stop_handler:
	push  hl
	ld    hl,msx_midi_handlers_stop
	pop   bc
	call  msx_midi_del_handler
	ret
	
	;; --------------------------------------------------------------------
	;; msx_midi_set_midi_in_handler - set custom handler for midi in
	;; --------------------------------------------------------------------
	;; in:	[hl] = address of midi in handler
	;; out:	-
	;; --------------------------------------------------------------------
	
msx_midi_set_midi_in_handler:
	ld    (msx_midi_handler_midi_in),hl
	ret

	;; --------------------------------------------------------------------
	;; msx_midi_clear_midi_in_handler - remove midi in handler
	;; --------------------------------------------------------------------
	;; in:	-
	;; out:	-
	;; --------------------------------------------------------------------
	
msx_midi_clear_midi_in_handler:
	ld   hl,0
	ld   (msx_midi_handler_midi_in),hl
	ret

	;; --------------------------------------------------------------------
	;; msx_midi_midi_out - put midi data on the midi out bus
	;; --------------------------------------------------------------------
	;; int:	[hl] = pointer to midi message, buffer should be filled as:
	;;             (hl) = length of message (16bits)
	;;             (hl+2) = midi messagae data
	;; out: -	
	;; --------------------------------------------------------------------
	
msx_midi_midi_out:
	ld   c,(hl)
	inc  hl
	ld   b,(hl)
	inc  hl
_msx_midi_midi_out0:
	push bc
	push hl
	ld   a,(hl)
	cp   $f8
	jr   nc,_msx_midi_midi_out_realtime
	;; if midi out buffer is full; enter synchronous mode
	push af
_msx_midi_midi_out2:
	ld   a,(msx_midi_out_buffer_count)
	cp   255
	jr   c,_msx_midi_midi_out3
	call msx_midi_midi_out_write
	jr   _msx_midi_midi_out2
_msx_midi_midi_out3:
	pop  af
	;; add to message buffer
	ld   hl,(msx_midi_out_buffer_wr_pointer)
	ld   (hl),a
	inc  hl
	ld   a,(msx_midi_out_buffer_count)
	inc  a
	ld   (msx_midi_out_buffer_count),a
	ld   (msx_midi_out_buffer_wr_pointer),hl
	call msx_midi_midi_out_write
	jr   _msx_midi_midi_out1
	;; handle realtime midi out data
_msx_midi_midi_out_realtime:
	;; if midi out buffer is full; enter synchronous mode
	push af
_msx_midi_midi_out_realtime0:
	ld   a,(msx_midi_out_buffer_rt_count)
	cp   255
	jr   c,_msx_midi_midi_out_realtime1
	call msx_midi_midi_out_write
	jr   _msx_midi_midi_out_realtime0
_msx_midi_midi_out_realtime1:
	pop  af
	;; add to message buffer
	ld   hl,(msx_midi_out_buffer_rt_wr_pointer)
	ld   (hl),a
	inc  hl
	ld   a,(msx_midi_out_buffer_rt_count)
	inc  a
	ld   (msx_midi_out_buffer_rt_count),a
	ld   (msx_midi_out_buffer_rt_wr_pointer),hl
	call msx_midi_midi_out_write
_msx_midi_midi_out1:
	pop  hl
	pop  bc
	inc  hl
	dec  bc
	ld   a,b
	or   c
	jr   nz,_msx_midi_midi_out0
	ret

	;; --------------------------------------------------------------------
	;; msx_midi_loop - process midi buffer
	;; --------------------------------------------------------------------
	;; in:	-
	;; out:	-
	;; --------------------------------------------------------------------

msx_midi_loop:
	;; write pending midi out
	ld   a,(msx_midi_out_buffer_count)
	or   a
	call nz,msx_midi_midi_out_write	
	;; handle interrupts
	di
	ld   a,(msx_midi_interrupt_flag)
	or   a
	ld   de,0
	ld   a,e
	ld   (msx_midi_interrupt_flag),a
	ei
	ld   hl,msx_midi_handlers_clock
 	call nz,msx_midi_handler_dispatch
	;; is there data available in the buffer
	ld   a,(msx_midi_in_buffer_count)
	or   a
	ret  z
	;; are we reading a message?
	ld   a,(msx_midi_message_count)
	or   a
	jp   z,_msx_midi_loop_new_message
	;; store midi data to buffer
	di
	ld   hl,(msx_midi_message_pointer)
	call _msx_midi_loop_read_buffer
	call _msx_midi_loop_check_buffer
	ei
	;; interrupted by a system message?
	cp   $f0
	jp   nc,_msx_midi_loop_system_message
_msx_midi_loop_data:
	;; nope, this is expected data, store.
	ld   (hl),a
	inc  hl
	ld   (msx_midi_message_pointer),hl
	;; all expected data received?
	ld   a,(msx_midi_message_count)
	dec  a
	ld   (msx_midi_message_count),a
	ret  nz
	;; dispatch midi message to midi handlers
	;; load midi message data in [bc]
	ld   bc,(msx_midi_message_buffer+1)
	ld   a,(msx_midi_message_buffer)
	;; store midi channel in [e]
	push af
	and  $0f
	inc  a
	ld   e,a
	ld   d,0
	pop  af
	;; dispatch note off messages - type = midi channel[e] + 0[d]
	ld   hl,msx_midi_handlers_noteoff
	cp   $90
	jp   c,msx_midi_handler_dispatch
	;; dispatch note on messages - type = midi channel[e] + 0[d]
	cp   $a0
	jr   c,_msx_midi_loop_dispatch_noteon ; handle velocity 0 => note off
	;; dispatch aftertouch messages - type = midi channel[e] + 0[d]
	ld   hl,msx_midi_handlers_aftertouch
	cp   $b0
	jp   c,msx_midi_handler_dispatch
	;; dispatch note cc message - type = midi channel[e] + cc[d]
	ld   d,c
	ld   hl,msx_midi_handlers_cc
	cp   $c0
	jp   c,msx_midi_handler_dispatch
	;; programm change - type = midi channel[e]
	ld   hl,msx_midi_handlers_programm_change
	cp   $d0
	jp   c,msx_midi_handler_dispatch
	;; channel pressure - type = midi channel[e]
	ld   hl,msx_midi_handlers_channel_pressure
	cp   $e0
	jp   c,msx_midi_handler_dispatch
	;; pith wheel - type = midi channel[e]
	cp   $f0
	ret  nc
	;; convert pitchwheel data (2*7bits to 1*14bits)
	ld   l,b
	ld   h,0
	add  hl,hl
	add  hl,hl
	add  hl,hl
	add  hl,hl
	add  hl,hl
	add  hl,hl
	add  hl,hl
	ld   a,l
	or   c
	ld   c,a
	ld   b,h
	ld   hl,msx_midi_handlers_pitchwheel
	jp   msx_midi_handler_dispatch
	;; note on; if velocity = 0, make it a note off message
_msx_midi_loop_dispatch_noteon:
	push af
	ld   a,b
	or   a			; velocity 0?
	jr   nz,_msx_midi_loop_dispatch_noteon0
	pop  af
	sub  %10		; make note off
	ld   (msx_midi_message_buffer),a
	ld   hl,msx_midi_handlers_noteoff
	jp   msx_midi_handler_dispatch
_msx_midi_loop_dispatch_noteon0:
	pop  af
	ld   hl,msx_midi_handlers_noteon
	jp   msx_midi_handler_dispatch
	;; check new midi message
_msx_midi_loop_new_message:
	ld   hl,msx_midi_message_buffer
	;; read data
	di
	call _msx_midi_loop_read_buffer
	call _msx_midi_loop_check_buffer
	ei
	;; is it a (unexpected) data byte?
	cp   $80
	jr   nc,_msx_midi_loop_new_message0 ; nope, this is a status byte
	ld   b,a
	;; check if in running status; otherwise ignore unexpected data!
	ld   a,(msx_midi_running_status)
	or   a
	ret  z
	;; running status; first process message byte
	push bc
	call _msx_midi_loop_new_message0
	pop  af
	;; then increase message pointer and handle it as regular data
	ld   hl,(msx_midi_message_pointer)
	jp   _msx_midi_loop_data
_msx_midi_loop_new_message0:
	ld   (hl),a
	;; don't let realtime messages influence the running status
	cp   $f8
	jr   nc,_msx_midi_loop_realtime_message
	;; regular message
	ld   (msx_midi_running_status),a
	;; $80-$8f = note off
;; 	cp   $90
;; 	jr   c,_msx_midi_loop_2byte_message
	;; $90-$9f = note on
;; 	cp   $a0
;; 	jr   c,_msx_midi_loop_2byte_message
	;; $a0-$af = aftertouch
;; 	cp   $b0
;; 	jr   c,_msx_midi_loop_2byte_message
	;; $b0-$bf = cc message
	cp   $c0
	jr   c,_msx_midi_loop_2byte_message
	;; $c0-$cf = program change
	cp   $d0
	jr   c,_msx_midi_loop_1byte_message
	;; $d0-$df = channel pressure
 	cp   $e0
 	jr   c,_msx_midi_loop_1byte_message
	;; $e0-$ef = pitch wheel
	cp   $f0
	jr   c,_msx_midi_loop_2byte_message
_msx_midi_loop_system_message:
	;; system common messages cancel any running status
	xor  a
	ld   (msx_midi_running_status),a
	;; $f0/$f7 = system exclusive
	;; $f1     = quarter frame message
	;; $f2     = song position pointer
	;; $f3     = song select
	;; $f6     = tune request
	ret
_msx_midi_loop_realtime_message:
	;; $f8     = midi clock (implemented in interrupt handler)
	;; $fa     = midi start
	ld   hl,msx_midi_handlers_start
	cp   $fa
	jp   z,msx_midi_handler_dispatch
	;; $fb     = midi continue
	;; $fc     = midi stop
	ld   hl,msx_midi_handlers_stop
	cp   $fc
	jp   z,msx_midi_handler_dispatch
	;; $fe     = active sense
	;; $ff     = midi reset
	ret

_msx_midi_loop_1byte_message:
	;; init a one byte message
	inc  hl
	ld   (msx_midi_message_pointer),hl
	ld   a,1
	ld   (msx_midi_message_count),a
	ret

_msx_midi_loop_2byte_message:
	;; init a two byte message
	inc  hl
	ld   (msx_midi_message_pointer),hl
	ld   a,2
	ld   (msx_midi_message_count),a
	ret

_msx_midi_loop_check_buffer:
	;; check if this is the last byte in the buffer, if so, reset pointers
	ld   c,a
	ld   a,(msx_midi_in_buffer_count)
	dec  a
	ld   (msx_midi_in_buffer_count),a
	ld   a,c
	ret  nz
	push hl
	ld   hl,msx_midi_in_buffer
	ld   (msx_midi_in_buffer_wr_pointer),hl
	ld   (msx_midi_in_buffer_rd_pointer),hl
	pop  hl
	ret

_msx_midi_loop_read_buffer:
	;; read midi data from buffer
	ld   de,(msx_midi_in_buffer_rd_pointer)
	ld   a,(de)
	inc  de
	ld   (msx_midi_in_buffer_rd_pointer),de
	ld   e,a
	;; is there a midi in handler we should call
	push hl
	ld   hl,(msx_midi_handler_midi_in)
	ld   a,h
	or   l
	ld   a,e
	jr   nz,_msx_midi_loop_read_buffer0
	pop  hl
	ret
_msx_midi_loop_read_buffer0:
	;; call midi read handler......
	push de
	push af
	ld   de,_msx_midi_loop_read_buffer1
	push de
	jp   (hl)
_msx_midi_loop_read_buffer1:
	pop  af
	pop  de
	pop  hl
	ret
		
	;; --------------------------------------------------------------------
	;; interrupt handlers
	;; --------------------------------------------------------------------

msx_midi_interrupt:
	di
	in   a,(msx_midi_status)
	ld   c,a
 	and  %10000000
	jr   z,_msx_midi_interrupt1
	ld   a,c
	and  %00000001 ; read buffer contains data
	jr   nz,_msx_midi_interrupt0
	ld   a,c
	and  %01110000
	jr   z,_msx_midi_interrupt0
	;; lost sync, clear running status seems to be the best choice
	xor  a
	ld   (msx_midi_running_status),a
_msx_midi_interrupt0:
	in   a,(msx_midi_datain)
	cp   $f8		; midi clock event?
	jr   z,_msx_midi_clock_ext
	ld   hl,(msx_midi_in_buffer_wr_pointer)
	ld   (hl),a
	inc  hl
	ld   a,(msx_midi_in_buffer_count)
	inc  a
	ret  z			; prevent buffer overflow (buffer = 256 bytes)
	ld   (msx_midi_in_buffer_count),a
	ld   (msx_midi_in_buffer_wr_pointer),hl
	ret

_msx_midi_interrupt1:
	in   a,($99)
	and  a
	ret  p
	;; vblank interrupt, use this as internal clock
	ld   a,(msx_midi_clock_type)
	or   a
	ret  nz
	ld   a,-1
	ld   (msx_midi_interrupt_flag),a
	ret

_msx_midi_clock_ext:
	ld   a,(msx_midi_clock_type)
	or   a
	ret  z
	ld   a,-1
	ld   (msx_midi_interrupt_flag),a
	ret

	;; --------------------------------------------------------------------
	;; msx_midi_midi_out_write - write midi out buffer to midi out port
	;; --------------------------------------------------------------------
	;; in:	-
	;; out:	-
	;; --------------------------------------------------------------------

msx_midi_midi_out_write:
	in   a,(msx_midi_status)
	and  %00000010
	call nz,_msx_midi_midi_out_write_rt
	in   a,(msx_midi_status)
	and  %00000010
	ret  z
_msx_midi_midi_out_write:
	;; write regular midi data
	ld   a,(msx_midi_out_buffer_count)
	or   a
	ret  z
	dec  a
	ld   (msx_midi_out_buffer_count),a
	push af
	ld   hl,(msx_midi_out_buffer_rd_pointer)
	ld   a,(hl)
	out  (msx_midi_dataout),a
	inc  hl
	ld   (msx_midi_out_buffer_rd_pointer),hl
	pop  af
	ret  nz
	ld   hl,msx_midi_out_buffer
	ld   (msx_midi_out_buffer_wr_pointer),hl
	ld   (msx_midi_out_buffer_rd_pointer),hl
	ret
_msx_midi_midi_out_write_rt
	;; handle realtime messages (these have priority 1)
	ld   a,(msx_midi_out_buffer_rt_count)
	or   a
	ret  z
	dec  a
	ld   (msx_midi_out_buffer_rt_count),a
	push af
	ld   hl,(msx_midi_out_buffer_rt_rd_pointer)
	ld   a,(hl)
	out  (msx_midi_dataout),a
	inc  hl
	ld   (msx_midi_out_buffer_rt_rd_pointer),hl
	pop  af
	ret  nz
	ld   hl,msx_midi_out_buffer_rt
	ld   (msx_midi_out_buffer_rt_wr_pointer),hl
	ld   (msx_midi_out_buffer_rt_rd_pointer),hl
	ret

	;; --------------------------------------------------------------------
	;; msx_midi_midi_reset - midi reset
	;; --------------------------------------------------------------------
	;; in:	-
	;; out:	-
	;; --------------------------------------------------------------------

msx_midi_midi_reset:
	di
	ld   a,%00000011 	; reset acia
	out  (msx_midi_command),a
	ret

	;; --------------------------------------------------------------------
	;; msx_midi_midi_init  -  init midi with interrupt
	;; --------------------------------------------------------------------
	;; in:	-
	;; out:	-
	;; --------------------------------------------------------------------

msx_midi_midi_init:
	call msx_midi_midi_reset
	ld   a,%10010101	; 8 data, no parity, 1 stop, int, 31250 bps
	out  (msx_midi_command),a
	ret

	;; --------------------------------------------------------------------
	;; add handler to handler list
	;; --------------------------------------------------------------------
	;; in:	[hl] = address of handler list
	;;      [bc] = address of handler to be added
	;;      [de] = handler type
	;; out:	-
	;; --------------------------------------------------------------------

msx_midi_add_handler:
	push de
_msx_midi_add_handler0:
	inc  hl
	inc  hl
	ld   a,(hl)
	inc  hl
	ld   e,(hl)
	inc  hl
	or   e
	jr   nz,_msx_midi_add_handler0
	pop  de
	dec  hl
	dec  hl
	dec  hl
	dec  hl
	ld   (hl),e
	inc  hl
	ld   (hl),d
	inc  hl
	ld   (hl),c
	inc  hl
	ld   (hl),b
	inc  hl
	ld   (hl),0
	inc  hl
	ld   (hl),0
	inc  hl
	ld   (hl),0
	inc  hl
	ld   (hl),0
	ret

	;; -------------------------------------------------------------------
	;; remove interrupt
	;; -------------------------------------------------------------------
	;; in:	[hl] = address of handler list
	;; 	[bc] = address of handler to remove
	;; -------------------------------------------------------------------
		
msx_midi_del_handler:
	ex   de,hl		; keeping the interface consistent
_msx_midi_del_handler0:	
	inc  de
	inc  de
	ld   a,(de)
	inc  de
	ld   l,a
	ld   a,(de)
	inc  de
	or   l
	ret  z
	call msx_midi_cphlbc
	jr   nz,_msx_midi_del_handler0
	push de
	pop  hl
	dec  de
	dec  de
	dec  de
	dec  de
_msx_midi_del_handler1:
	ldi
	ldi
	ld   a,(hl)
	ldi
	ld   c,(hl)
	or   c
	ldi
	jr   nz,_msx_midi_del_handler1
	ret

	;; -------------------------------------------------------------------
	;; msx_midi_handler_dispatch - call all handlers on handler list
	;; -------------------------------------------------------------------
	;; in:	[hl] = address of handler list
	;;      [bc] = optional data to throw to handler methods
	;;      [de] = handler type ([d]=extra,[e]=midi channel)
	;; out:	-
	;; -------------------------------------------------------------------
	
msx_midi_handler_dispatch:
	;; check handler method equals type
	ld   a,(hl)		; midi channel
	or   a			; 0 = dispatch all midi channels to me
	inc  hl
	jr   z,_msx_midi_handler_dispatch0
	cp   e			; check midi channel with handler midi channel
	jr   z,_msx_midi_handler_dispatch0
	inc  hl
	jr   _msx_midi_handler_dispatch1
_msx_midi_handler_dispatch0:
	ld   a,(hl)
	inc  hl
	cp   d			; check additional type data
	jr   z,_msx_midi_handler_dispatch2
	;; check next handler in list
_msx_midi_handler_dispatch1:
	ld   a,(hl)
	push bc
	ld   c,a
	inc  hl
	ld   a,(hl)
	inc  hl
	or   c
	pop  bc
	ret  z
	jr   msx_midi_handler_dispatch
	;; call actual handler method
_msx_midi_handler_dispatch2:
	push de
	ld   d,e
	push de			; d = midi channel
	ld   a,(hl)
	ld   e,a
	inc  hl
	ld   d,(hl)
	inc  hl
	or   e
	jr   z,_msx_midi_handler_done
	pop  af			; a = midi channel, bc = additional data
	push hl
	ex   de,hl
	ld   de,_msx_midi_handler_callback
	push de
	jp   (hl)
_msx_midi_handler_callback:
	pop  hl
	pop  de
	jr   msx_midi_handler_dispatch
_msx_midi_handler_done:
	pop  af
	pop  de
	ret

	;; -------------------------------------------------------------------
	;; compare hl with bc
	;; -------------------------------------------------------------------
	;; in:	[hl] = value 1
	;;      [bc] = value 2
	;; out:	[ f] = result of comparison between value 1 and value 2
	;; -------------------------------------------------------------------
	
msx_midi_cphlbc:
	ld   a,h
	sub  b
	ret  nz
	ld   a,l
	sub  c
	ret
